/* FILE: standarddriver.cc            -*-Mode: c++-*-
 *
 * Example concrete Oxs_Driver class.
 *
 */

#include "nb.h"
#include "standarddriver.h"
#include "director.h"
#include "simstate.h"
#include "evolver.h"
#include "key.h"
#include "mesh.h"
#include "scalarfieldinit.h"
#include "vectorfieldinit.h"
#include "energy.h"		// Needed to make MSVC++ 5 happy

// Oxs_Ext registration support
OXS_EXT_REGISTER(Oxs_StandardDriver);

/* End includes */

#define BUF_SIZE 1024
static char buf[BUF_SIZE];	// Buffer for building interpreter results

// Constructor
#define OSO_INIT(name,descript,units) \
   name##_output(this,InstanceName(),descript,units,0, \
           &Oxs_StandardDriver::Fill__##name##_output,newdtr)
Oxs_StandardDriver::Oxs_StandardDriver(
  const char* name,     // Child instance id
  Oxs_Director* newdtr, // App director
  Tcl_Interp* safe_interp, // Safe interpreter
  const char* argstr)   // MIF input block parameters
  : Oxs_Driver(name,newdtr,safe_interp,argstr),
    OSO_INIT(last_timestep,"Last time step","s"),
    OSO_INIT(elapsed_time,"Elapsed time","s"),
    OSO_INIT(max_dm_dt,"Max dm/dt","deg/ns"),
    OSO_INIT(total_energy,"Total energy","J"),
    OSO_INIT(iteration_count,"Iteration",""),
    OSO_INIT(stage_iteration_count,"Stage iteration",""),
    OSO_INIT(stage_number,"Stage",""),
    magnetization_output(this,InstanceName(),"Magnetization","A/m",1,
			 &Oxs_StandardDriver::Fill__magnetization_output,
			 newdtr)
{
  // Process arguments
  max_timestep=GetRealInitValue("max_timestep");
  if(max_timestep<=0.0) {
    char buf[4096];
    Oc_Snprintf(buf,sizeof(buf),
		"Invalid MIF input block detected for object %s:"
		" Specified max time step is %g",InstanceName(),
		max_timestep);
    throw Oxs_Ext::Error(buf);
  }

  min_timestep=GetRealInitValue("min_timestep");

  stopping_dm_dt=GetRealInitValue("stopping_dm_dt");

  const vector<string>*  vsptr; // Temp pointer
  number_of_stages=1;  // Default value: 1 stage
  vsptr = FindInitValue("number_of_stages");
  if(vsptr!=NULL) {
    number_of_stages=GetIntInitValue("number_of_stages");
  }

  stage_iteration_limit=0;  // Default value: no limit
  vsptr = FindInitValue("stage_iteration_limit");
  if(vsptr!=NULL) {
    stage_iteration_limit=GetIntInitValue("stage_iteration_limit");
  }

  total_iteration_limit=0;  // Default value: no limit
  vsptr = FindInitValue("total_iteration_limit");
  if(vsptr!=NULL) {
    total_iteration_limit=GetIntInitValue("total_iteration_limit");
  }
 
  CheckInitValueParamCount("mesh",1);
  Oxs_Mesh* mesh = dynamic_cast<Oxs_Mesh*>
    (director->FindExtObject((*FindInitValue("mesh"))[0]));
  if(mesh==NULL) {
    string msg="Unable to identify Oxs_Mesh object with name "
      + (*FindInitValue("mesh"))[0];
    throw Oxs_Ext::Error(msg.c_str());
  }
  DeleteInitValue("mesh");
  meshkey.Set(mesh);  // Sets a dep lock

  auto_ptr<Oxs_ScalarFieldInit>
    Msinit(NewScalarFieldInit(safe_interp,"Ms"));

  auto_ptr<Oxs_VectorFieldInit> foo(NewVectorFieldInit(safe_interp,"m0"));
  m0 = foo;
  /// Stroustrup specifies an auto_ptr .reset() member function,
  /// but that is apparently not supported by some common STL's.
  /// This above creation of a temporary (foo) is a workaround.

  VerifyAllInitArgsUsed();

  // Fill Ms and Ms_inverse array
  Msinit->FillMeshValue(mesh,Ms);
  Ms_inverse.AdjustSize(mesh);
  for(UINT4m i=0;i<mesh->Size();i++) {
    if(Ms[i]!=0.0) Ms_inverse[i]=1.0/Ms[i];
    else           Ms_inverse[i]=0.0;
  }

  // Reserve space for initial state (see GetInitialState() below)
  director->ReserveSimulationStateRequest(1);
}

Oxs_ConstKey<Oxs_SimState>
Oxs_StandardDriver::GetInitialState() const
{
  Oxs_Key<Oxs_SimState> initial_state;
  director->GetNewSimulationState(initial_state);
  Oxs_SimState& istate = initial_state.GetWriteReference();
  istate.last_timestep   =  0.0;
  istate.elapsed_time    =  0.0;
  istate.max_dm_dt       = -1.0;     // Dummy value.
  istate.total_energy    =  0.0;     // Dummy value.
  istate.iteration_count =  0;
  istate.stage_iteration_count =  0;
  istate.stage_number    =  0;

  istate.mesh = meshkey.GetPtr();

  istate.Ms = &Ms;
  istate.Ms_inverse = &Ms_inverse;
  m0->FillMeshValue(istate.mesh,istate.spin);

  istate.valid_state=0;  // Technically, this is not a valid
  /// state because max_dm_dt and total_energy aren't set.
  return initial_state;
}


BOOL Oxs_StandardDriver::Init()
{ 
  Init_Driver();  // Run init routine in parent.  This will call
  return 1;       // Oxs_StandardDriver::GetInitialState()
}

Oxs_StandardDriver::~Oxs_StandardDriver()
{}

BOOL
Oxs_StandardDriver::StageDone(const Oxs_SimState& state, string& result) const
{ // Note: The iteration counts are effective 1-based, because
  // the original state in each stage is marked 'invalid_state,'
  // since max_dm_dt & energy aren't set.  Thus, since state
  // iteration counts record the count for a completed state,
  // when the iteration count is the same as the limit, the limit
  // is reached.
  if( stage_iteration_limit>0
      && state.stage_iteration_count >= stage_iteration_limit) goto done;
  if( total_iteration_limit>0
      && state.iteration_count >= total_iteration_limit ) goto done;
  if( !state.valid_state ) return 0;
  if( state.max_dm_dt <= stopping_dm_dt ) goto done;
  return 0;
done:
  result.append("Oc_EventHandler Generate Oxs Stage\n");
  return 1;
}

BOOL Oxs_StandardDriver::Done(const Oxs_SimState& state, string& result) const
{ // Note: See note on iteration counts in StageDone() above.
  if( total_iteration_limit>0
      && state.iteration_count >= total_iteration_limit ) goto done;
  if( number_of_stages >0
      && state.stage_number >= number_of_stages ) goto done;
  return 0;
done:
  result.append("Oc_EventHandler Generate Oxs Done\n");
  return 1;
}

void
Oxs_StandardDriver::NewState(const Oxs_SimState& state, string& result) const
{
  Oc_Snprintf(buf, sizeof(buf), 
              "Oc_EventHandler Generate Oxs Step -step %d -stage %d\n",
              state.iteration_count, state.stage_number);
  result.append(buf);
}

void Oxs_StandardDriver::FillStatePartA
(const Oxs_SimState& old_state,
 Oxs_SimState& new_state) const
{
  new_state.valid_state  = 0;
  new_state.stage_number = old_state.stage_number;
  new_state.Ms           = old_state.Ms;
  new_state.mesh         = old_state.mesh;
  new_state.spin.AdjustSize(new_state.mesh);
}

void Oxs_StandardDriver::FillStatePartB(Oxs_SimState& work_state) const
{
  if(work_state.last_timestep<min_timestep)
    work_state.last_timestep=min_timestep;
  if(work_state.last_timestep>max_timestep)
    work_state.last_timestep=max_timestep;
}

////////////////////////////////////////////////////////////////////////
// State-based outputs, maintained by the driver.  These are
// conceptually public, but are specified private to force
// clients to use the output_map interface in Oxs_Director.

#define OSO_FUNC(NAME) \
void Oxs_StandardDriver::Fill__##NAME##_output(const Oxs_SimState& state) \
{ NAME##_output.cache.state_id=state.Id(); \
  NAME##_output.cache.value=state.NAME; }

OSO_FUNC(last_timestep)
OSO_FUNC(elapsed_time)
OSO_FUNC(total_energy)
OSO_FUNC(iteration_count)
OSO_FUNC(stage_iteration_count)
OSO_FUNC(stage_number)

void Oxs_StandardDriver::Fill__max_dm_dt_output(const Oxs_SimState& state)
{
  max_dm_dt_output.cache.state_id=state.Id();
  max_dm_dt_output.cache.value = state.max_dm_dt * 180e-9 / PI;
  /// Convert from radians/second to deg/ns
}

void
Oxs_StandardDriver::Fill__magnetization_output(const Oxs_SimState& state)
{
  magnetization_output.cache.state_id=0;
  magnetization_output.cache.value.AdjustSize(state.mesh);
  UINT4m size=state.mesh->Size();
  const Oxs_MeshValue<ThreeVector>& spin = state.spin;
  const Oxs_MeshValue<REAL8m>& Ms = *(state.Ms);
  Oxs_MeshValue<ThreeVector>& mag = magnetization_output.cache.value;
  for(UINT4m i=0;i<size;i++) {
    mag[i] = spin[i];
    mag[i] *= Ms[i];
  }
  magnetization_output.cache.state_id=state.Id();
}
