/* FILE: uuanisotropy.cc            -*-Mode: c++-*-
 *
 * Uniform Uniaxial Anisotropy, derived from Oxs_Energy class.
 *
 */

#include "oc.h"
#include "nb.h"
#include "threevector.h"
#include "director.h"
#include "simstate.h"
#include "ext.h"
#include "key.h"
#include "mesh.h"
#include "meshvalue.h"
#include "uniaxialanisotropy.h"
#include "energy.h"		// Needed to make MSVC++ 5 happy

// Oxs_Ext registration support
OXS_EXT_REGISTER(Oxs_UniaxialAnisotropy);

/* End includes */


// Constructor
Oxs_UniaxialAnisotropy::Oxs_UniaxialAnisotropy(
  const char* name,     // Child instance id
  Oxs_Director* newdtr, // App director
  Tcl_Interp* safe_interp, // Safe interpreter
  const char* argstr)   // MIF input block parameters
  : Oxs_Energy(name,newdtr,safe_interp,argstr), mesh_id(0)
{
  // Process arguments
  // Note: Stroustrup specifies an auto_ptr .reset() member function,
  // but that is apparently not supported by some common STL's.  As
  // a workaround, create a temporary auto_ptr, and use operator=().
  auto_ptr<Oxs_ScalarFieldInit>
    foo(NewScalarFieldInit(safe_interp,"K1"));
  K1_init = foo;
  auto_ptr<Oxs_VectorFieldInit>
    bar(NewVectorFieldInit(safe_interp,"axis"));
  axis_init = bar;
  VerifyAllInitArgsUsed();
}

BOOL Oxs_UniaxialAnisotropy::Init()
{ return 1; }

void Oxs_UniaxialAnisotropy::GetEnergyAndField
(const Oxs_SimState& state,
 Oxs_MeshValue<REAL8m>& energy,
 Oxs_MeshValue<ThreeVector>& field
 ) const
{
  const Oxs_MeshValue<REAL8m>& Ms_inverse = *(state.Ms_inverse);
  const Oxs_MeshValue<ThreeVector>& spin = state.spin;
  UINT4m size = state.mesh->Size();
  if(mesh_id != state.mesh->Id()) {
    // This is either the first pass through, or else mesh
    // has changed.
    mesh_id = 0;
    K1_init->FillMeshValue(state.mesh,K1);
    axis_init->FillMeshValue(state.mesh,axis);
    for(UINT4m i=0;i<size;i++) {
      // Check that axis is a unit vector:
      const REAL8m eps = 1e-14;
      if(fabs(axis[i].MagSq()-1)>eps) {
	string msg="Invalid initialization detected for object "
	  + string(InstanceName())
	  + ": Anisotropy axis isn't norm 1";
	throw Oxs_Ext::Error(msg.c_str());
      }
    }
    mesh_id = state.mesh->Id();
  }

  for(UINT4m i=0;i<size;++i) {
    REAL8m k = K1[i];
    REAL8m field_mult = (2.0/MU0)*k*Ms_inverse[i];
    if(field_mult==0.0) {
      energy[i]=0.0;
      field[i].Set(0.,0.,0.);
      continue;
    }
    REAL8m dot = axis[i]*spin[i];
    field[i] = (field_mult*dot) * axis[i];
    if(k>0) {
      energy[i] = -k*(dot*dot-1.0); // Make easy axis zero energy
    } else {
      energy[i] = -k*dot*dot; // Easy plane is zero energy
    }
  }
}



